/************************************************************************

          Pinc CModel Header File

History
=======
20140221 SFP007 Added EOL and IGN characters.
20130816 SFP006 Added MinGW64
20130716 ATH005 Added Linux64
20120610 ATH004 Added Linux
     21/11/90  Rod Fixed SIGNED Problem on KNR Compilers
     12/ 4/90  Rod Added OS9 Support
      5/ 4/90  Rod Added Support For Small Model
     14/11/89  Author: Rod Crawford

This file describes the type of C Model you are compiling Pinc
on - Large/Small. In addition it also defines the C Compiler
you are using for any implementation specific parts of Pinc.

************************************************************************/

#define MicrosoftC 1
#define SUN 2
#define OS9 3
#define TurboC 4
#define RISCOS 5
#define Linux 6
#define Linux64 7
#define MinGW64 8

/*
  Select the required implementation here
*/
// #define IMPLEMENTATION MicrosoftC
// #define IMPLEMENTATION Linux
// #define IMPLEMENTATION Linux64
#define IMPLEMENTATION MinGW64

/*
Some #defines control how the C compiler is used. You may need to
write some test code (see the tests directory) to find out the
actual sizes of short, int, long and long long for your compiler.

#define CELL_SIZE (CELL) (sizeof(long long))
Define the size of a Forth cell in address units (bytes).

#define CELL long long
Define the C type used for a cell.

#define CELL_IS_LONG
If a cell is a LONG, define this.

#define CELL_IS_LONG_LONG
If a cell is a LONG LONG, define this.

#define NO_2CELL_TYPE
If there is no type large enough for two cells, define
this.

#define TWO_CELL long long
If there is a type large enough for two cells, define it
*/

/*
These #defines control how the Forth operates.

These two define how EXPECT and ACCEPT work. They allow Forth apps
to know which character of LF and CR is treated as the proper end
of line character.

#define EOLCHAR_CON 13		// console input line terminator
#define IGNCHAR_CON 10		// ignore this character
*/

#if IMPLEMENTATION == MinGW64
// 64 bit Windows app
// #warning " === MINGW64 ==="
#define CELL_SIZE (CELL) (sizeof(long long))
#define CELL long long
#define CELL_IS_LONG_LONG
#define NO_2CELL_TYPE
#define EOLCHAR_CON 13		// console input line terminator
#define IGNCHAR_CON 10		// ignore this character
#endif

#if IMPLEMENTATION == Linux64
// #define KNR
#warning "LINUX ===== 64"
#define CELL_SIZE (CELL) (sizeof(long))
#define CELL long
#define CELL_IS_LONG
#define EOLCHAR_CON 10		// console input line terminator
#define IGNCHAR_CON 13		// ignore this character
#endif

#if IMPLEMENTATION == SUN
#define KNR
#define void
#define CELL_SIZE (CELL) 4
#define CELL int
#define EOLCHAR_CON 10		// console input line terminator
#define IGNCHAR_CON 13		// ignore this character
#endif

#if IMPLEMENTATION == Linux
// #define KNR
#warning "LINUX"
#define CELL_SIZE (CELL) 4
#define CELL int
#define EOLCHAR_CON 10		// console input line terminator
#define IGNCHAR_CON 13		// ignore this character
#endif

#if IMPLEMENTATION == MicrosoftC
#define EOLCHAR_CON 13		// console input line terminator
#define IGNCHAR_CON 10		// ignore this character

/* Uncomment the following block if you are using a Small Model
   And Comment out the one below it          */
/*
#define huge
#define CELL_SIZE (CELL) 2
#define CELL int
*/

/* Uncomment the block below if you are using a Large Model
   And Coment out the one above          */
#define CELL_SIZE (CELL) 4
#define CELL long
#define CELL_IS_LONG

#endif		/* MicrosoftC */

#if IMPLEMENTATION == OS9
#define KNR
#define sp psp
#define rp rsp
#define void
#define CELL_SIZE (CELL) 4
#define CELL int
#define EOLCHAR_CON 10		// console input line terminator
#define IGNCHAR_CON 13		// ignore this character
#endif

#if IMPLEMENTATION == TurboC
#define EOLCHAR_CON 13		// console input line terminator
#define IGNCHAR_CON 10		// ignore this character

/* Uncomment the following block if you are using a Small Model
   And Comment out the one below it          */
/*
#define huge
#define CELL_SIZE (CELL) 2
#define CELL int
*/
#endif

/* Signed is an ANSI Extension so... */
#ifdef KNR
#define signed
#endif

#define NBITS (CELL_SIZE * 8)	/* Number of bits in a cell */
#define HNBITS (NBITS / 2)	/* Half the number of bits in a cell. */

/* Assume twos-complement arithmetic to calculate lower half. */
#define LO_HALF(n) ( (n) & (((CELL)1 << HNBITS) - 1) )
#define HI_HALF(n) ( (n) >> HNBITS )

/* Maths macro */
#define DULT(du1l,du1h,du2l,du2h) ( (du2h<du1h) ? FALSE : ( (du2h==du1h) ? (du1l<du2l) : TRUE) )
