/**********************************************************************

        Pinc Error Handler

Author: Rod Crawford

Date: 15/11/89

Mods: 12/ 4/90 Rod Added OS9 Support

This file contains the Pinc Error Handler.


**********************************************************************/


#include "cmodel.h"
#include "pinc.h"
#include "lowram.h"
#include "strsizes.h"
#include "fname.h"
#include "vmsize.h"
#include "misc.h"


/**************************  external data *******************************/


#if IMPLEMENTATION == MicrosoftC
    extern unsigned char huge ram[ RAM_TOP ];
#else
#if IMPLEMENTATION == OS9
    extern remote unsigned char ram[ RAM_TOP ];
#else
    extern unsigned char ram[ RAM_TOP ];
#endif
#endif


extern unsigned CELL sp, rp, ip;
extern unsigned CELL compiling;


extern FILE  *error_message_stream;
extern struct from_fileblock {
          char filename[ FROM_FILENAME_LENGTH ];
          int page_number, line_number, line_in_page_number,
              last_page_required;
          FILE *stream;
            } input;



/******************* External Function Prototypes  ************************/

#ifdef KNR

#ifdef CHECK_STACKS
extern void push_data();
extern unsigned CELL pop_data();
#else
#include "stacks.h"

#endif
extern void init_from_stack();
extern void trash_rest_of_line();
#else

#ifdef CHECK_STACKS
extern void push_data(unsigned CELL data);
extern unsigned CELL pop_data(void );
#else
#include "stacks.h"
#endif

extern void init_from_stack(void );
extern void trash_rest_of_line(void );
#endif

#include "extproto.h"

/*********************** Function Prototypes  *****************************/

#ifdef KNR
void init_error_handler();
int error_message_file();
void do_error();
void show_this_error();
void display_error();
int warm_vector_set();
#else
void init_error_handler(void );
int error_message_file(void );
void do_error(unsigned CELL error_number);
void show_this_error(unsigned CELL error_number);
void display_error(char *error_message);
int warm_vector_set(void );
#endif


/***************************  Public Global Data  ***********************

error_filename  holds the name of the error file used by Pinc
                this is set to s default value but may be
                changed using the is-error-file word.
                This facility is used bye the cross compiler.

************************************************************************/

char error_filename[ ERROR_FILENAME_LENGTH ] = ERROR_FILE;




/***************************  init_error_handler  ***************************

Desc: The error message file is opened, if it exists. If it does not
      error_message stream will be equal to NULL.

**************************************************************************/

void init_error_handler()
{
    error_message_stream = fopen( error_filename, "r" );
}




/*************************  error_message_file  ************************

Desc: Returns true if there is an error message file, else false

Ret:  1 or 0.

**************************************************************************/

int error_message_file()
{
    return( error_message_stream != NULL );
}



/*********************************  do_error  ***************************

Desc: Prints the given error number and if possible the associated
      error message.
      If the warm start vector has been initialised then a warm start is
      performed.

************************************************************************/

 void do_error( error_number )
 unsigned CELL error_number;
 {
    show_this_error( error_number );
    if( warm_vector_set()) {
        sp = (CELL) -1;
        rp = (CELL) -1;
        push_data( (CELL) WARM_VECTOR );
        fetch();
        ip = pop_data();
    }
    else {
     compiling = 0;
     init_io();
     trash_rest_of_line();
    }
}




/*************************  show_this_error  ******************************

Desc: displays the error whose number is given.
      If no error mesage file is open then only the number is
      displayed.
      If the error number exceeds the number of errors in the file
      then a "no such error message is printed".

In:   error_number  number of the error to be displayed

***********************************************************************/


void show_this_error( error_number )
unsigned CELL error_number;
{
    char error_message_string[ LINE_LENGTH ], *error_message,
         error_number_string[ LINE_LENGTH ], *status;

#ifdef CELL_IS_LONG_LONG
    sprintf( error_number_string, "Error #%lld ", error_number );
#else
#ifdef CELL_IS_LONG
    sprintf( error_number_string, "Error #%ld ", error_number );
#else
    sprintf( error_number_string, "Error #%d ", error_number );
#endif
#endif
    error_message_string[0] = '\0';
    if( error_message_file()) {
        rewind( error_message_stream );
        {
            int i;

            for( i=0; i<=(int) error_number; i++ )
                status = fgets( error_message_string, LINE_LENGTH,
                                error_message_stream );
        }
    }
    error_message = strcat( error_number_string,
                                   error_message_string );
    display_error( error_message );
}




/****************************  display_error  ************************

Desc: Informs the user that an error has occurred.
      The line on which the error occurred is printed followed
      by a pointer where it occurred and the given error message.

In:   error_message  string to be printed as error message.

*******************************************************************/

void display_error( error_message )
char error_message[];
{
    int i;
    unsigned CELL *lineptr;

    lineptr = (unsigned CELL *) &ram[ TO_IN ];
    fprintf( stderr, "\n%s\n", &ram[ TIB ] );
    for( i=0; i<(*lineptr)-1; i++ )  fprintf( stderr, " " );
    fprintf( stderr, "^ %s\n", error_message );
    if( strcmp( input.filename, "Console" ))
     fprintf( stderr, "%s: Line %d    Page/Line %d/%d",
           input.filename, input.line_number, input.page_number,
           input.line_in_page_number );
    trash_rest_of_line();
    init_from_stack();
}





/***********************  warm_vector_set  ***************************

Desc: Returns true if the warm start vector has been
      setup.
      N.B. The warm start vector cannot be made to point to
           the cold start vector.
           (Why would you want to do that anyway?)

Ret: 1  if the vector has been set
     0  if it has not been set.

**********************************************************************/

int warm_vector_set()
{
    unsigned CELL *cell;

    cell = (unsigned CELL *) &ram[ WARM_VECTOR ];
    return(  (*cell) != (CELL) 0 );
}
