/**********************************************************************

            Pinc Head Traversal Functions.

Author: Rod Crawford

Date: 15/11/89
Mods:  9/ 4/90 Rod Added execution_token function token return either
             a secondaries cfa or the contents of that field
             Also had to add is_primitive macro to facilitate the
             above.
      12/ 4/90 Rod Added OS9 Support

The following functions convert given address fields to other address
fields.

**********************************************************************/

#include "cmodel.h"
#include "pinc.h"
#include "vmsize.h"
#include "nfmasks.h"


/**************************  external data *******************************/


#if IMPLEMENTATION == MicrosoftC
    extern unsigned char huge ram[ RAM_TOP ];
#else
#if IMPLEMENTATION == OS9
    extern remote unsigned char ram[ RAM_TOP ];
#else
    extern unsigned char ram[ RAM_TOP ];
#endif
#endif


extern unsigned CELL compiling;


#include "enum.h" /* Just to get prim_null_prim */

/*******************  External Function Prototypes  *************************/

#ifdef KNR
extern unsigned CELL cell_aligned();
#else
extern unsigned CELL cell_aligned(unsigned CELL address);
#endif

/***************************  Function Protoypes  ***********************/

#ifdef KNR
unsigned CELL nfa_to_lfa();
unsigned CELL lfa_to_nfa();
unsigned CELL nfa_to_cfa();
unsigned CELL cfa_to_nfa();
unsigned CELL cfa_to_voc_link();
unsigned CELL next_voc_link();
int is_immediate();
int smudged();
unsigned CELL execution_token();
#else
unsigned CELL nfa_to_lfa(unsigned CELL address);
unsigned CELL lfa_to_nfa(unsigned CELL address);
unsigned CELL nfa_to_cfa(unsigned CELL nfa);
unsigned CELL cfa_to_nfa(unsigned CELL cfa);
unsigned CELL cfa_to_voc_link(unsigned CELL cfa);
unsigned CELL next_voc_link(unsigned CELL this_voc_link);
int is_immediate(unsigned CELL nfa);
int smudged( unsigned CELL nfa );
unsigned CELL execution_token(unsigned CELL lfa);
#endif


/***************************  nfa_to_lfa  ****************************/

unsigned CELL nfa_to_lfa( address )
unsigned CELL address;
{
    return( address - CELL_SIZE );
}




/***************************  lfa_to_nfa  ****************************/

unsigned CELL lfa_to_nfa( address )
unsigned CELL address;
{
    return( address + CELL_SIZE );
}



/***************************  nfa_to_cfa  ****************************/


unsigned CELL nfa_to_cfa( nfa )
unsigned CELL nfa;
{
    return( cell_aligned( nfa + strlen( &ram[ nfa+1 ]) + 2 ));
}



/***************************  cfa_to_nfa  ****************************/


unsigned CELL cfa_to_nfa( cfa )
unsigned CELL cfa;
{
    while( !(ram[ --cfa ] & NFA_MASK ));
    return( cfa );
}


/*************************  cfa_to_voc_link  **************************/

unsigned CELL cfa_to_voc_link( cfa )
unsigned CELL cfa;
{
    return( cfa + CELL_SIZE );
}



/***********************  next_voc_link  *****************************/

unsigned CELL next_voc_link( this_voc_link )
unsigned CELL this_voc_link;
{
    unsigned CELL *cell;

    cell = (unsigned CELL *) &ram[ cfa_to_voc_link(this_voc_link) ];
    return( *cell );
}



/******************************  is_immediate  **************************

Desc: Returns true if the given nfa has the immediate bit set.

In:   nfa  given nfa to be checked.

Ret:  True if nfa bit set else False.

**********************************************************************/

int is_immediate( nfa )
unsigned CELL nfa;
{
    return( (ram[ nfa ] & IMMEDIATE_MASK) != 0);
}


/************************  smudged  *********************************

Desc: returns true if the latest word has been defined else false

Rets: True i.e. the actual smudge bit
      False i.e. 0


********************************************************************/

int smudged( nfa )
unsigned CELL nfa;
{
    return ( ram[ nfa ] & SMUDGE_MASK );
}



/************************  is_primitive ****************************

Desc: Returns True is the gicen instruction is within the
      numeric range of the enumeration representing Pinc primitive
      opcodes else false.
      Note For this funciton to operate correctly the enumeration
      must begin with the opcode noop and end with the prim_null_primitive.

      THIS IS DEFINED AS MACRO TO IMPORVE EXECUTION SPEED
      THE DEFINTION ALSO OCCURS IN INNER.C
      DON'T CHANGE ONE WITHOUT TOTHER!

In:  instruction  a number possibly representing a Pinc primitive
                  opcode.

*******************************************************************/

/*
int is_primitive( instruction )
unsigned CELL instruction;
{
    return( ( instruction >= (unsigned CELL) prim_noop )  &&
         ( instruction <= (unsigned CELL) prim_null_primitive ));
}
*/

#define is_primitive( instruction ) \
    ( instruction <= (unsigned CELL) prim_null_primitive )



/**********************  execution_token  **************************

Desc: Returns the exectution token for a given lfa.

In:   lfa  the link field address of the word in question.

Rets: Exectution token for the word

*********************************************************************/

unsigned CELL execution_token( lfa )
unsigned CELL lfa;
{
    unsigned CELL cfa_contents, cfa;

    cfa = nfa_to_cfa( lfa_to_nfa( lfa ));
    cfa_contents = *((unsigned CELL *) &ram[ cfa ]);

    if( !compiling && is_primitive( cfa_contents ))
     return( cfa );
    return( cfa_contents );
}
