/*********************************************************************

        C/CWeed Forth Inner Interpreter


Author: Rod Crawford

Date: 16/11/89
Mods:  5/ 4/90 Rod Replaced enum with #defs for K&R compatabilty
       6/ 4/90 sfp Revised the inner interpreter loop and removed
             one extra pointer dereference that the compiler
             ought to remove anyway.
      12/ 4/90 Rod Added OS9 support
      24/ 7/90 sfp Call to PUSH_RETURN in INNER_INTERPRETER replaced
             by in-line code. 68000 cross compiler execution
             changes from 161 seconds to 138 seconds.
             Definition of TOKEN in INNER_INTERPRETER made REGISTER
             to improve performance on larger (32-bit) machines.

This file contains the C/CWeed Forth inner interpreter.


***********************************************************************/

#include "cmodel.h"
#include "pinc.h"
#include "vmsize.h"
#include "misc.h"



/**************************  external data *******************************/



#if IMPLEMENTATION == MicrosoftC
    extern unsigned char huge ram[ RAM_TOP ];
#else
#if IMPLEMENTATION == OS9
    extern remote unsigned char ram[ RAM_TOP ];
#else
    extern unsigned char ram[ RAM_TOP ];
#endif
#endif


extern unsigned CELL ip, rp;
extern unsigned CELL return_stack[ RETURN_STACK_TOP ];

#ifdef CWEED
extern int kill_switch;
#endif
extern jmp_buf environment;


#include "enum.h"


typedef void (*function_pointer) (void);
extern function_pointer execution_vector[ prim_null_primitive ];








/******************* External Function Prototypes  ************************/

#ifdef KNR
extern void do_error();
extern int number();

#ifdef CHECK_STACKS
extern void push_data();
extern void push_return();
#else
#include "stacks.h"
#endif

extern unsigned CELL lfa_to_nfa();
extern unsigned CELL nfa_to_cfa();
#else
extern void do_error(unsigned CELL error_number);
extern int number(char *string,int string_ptr,unsigned CELL *value);

#ifdef CHECK_STACKS
extern void push_data(unsigned CELL data);
extern void push_return(unsigned CELL address);
#else
#include "stacks.h"
#endif

extern unsigned CELL lfa_to_nfa(unsigned CELL address);
extern unsigned CELL nfa_to_cfa(unsigned CELL nfa);
#endif


/**********************  Function Prototypes  *****************************/

#ifdef KNR
void inner_interpreter();
void execute_primitive();
void execute();
#else
void inner_interpreter(void );
void execute_primitive( int instruction);
void execute(unsigned CELL cfa);
#endif



/************************  is_primitive ****************************

Desc: Returns True is the gicen instruction is within the
      numeric range of the enumeration representing Pinc primitive
      opcodes else false.
      Note For this funciton to operate correctly the enumeration
      must begin with the opcode noop and end with the prim_null_primitive.

      THIS IS DEFINED AS MACRO TO IMPORVE EXECUTION SPEED
      THE DEFINTION ALSO OCCURS IN HEADTRAV.C
      DON'T CHANGE ONE WITHOUT TOTHER!

In:  instruction  a number possibly representing a Pinc primitive
                  opcode.

*******************************************************************/

/*
int is_primitive( instruction )
unsigned CELL instruction;
{
    return( ( instruction >= (unsigned CELL) prim_noop )  &&
         ( instruction <= (unsigned CELL) prim_null_primitive ));
}
*/

#define is_primitive( instruction ) \
    ( instruction <= (unsigned CELL) prim_null_primitive )



/***************************  inner_interpreter  ************************

Desc: Two versions of the inner interpreter are provided. One for
      CWEED Forth the other Pinc. This is necessary because when CWEED
      forth is running the actual text interpreter is a CFunction called
      cweed_interpret. This kicks off the inner interpreter to execute a
      word. In order to get back to the text interpreter from the inner
      interpreter a flag called the kill switch must be set. This is
      generally done by NEXT. When the switch has been set the inner
      interpreter exits back to cweed_interpret.
      In the case of full Pinc the outer interpreter is a proper forth
      word so at not time should we exit the inner interpret (except
      indirectly via a long jump from a bye or error condition).
      The two interpreters are described below:

      CWeed Forth Version.
      We continue to run the interpreter whilst the kill_switch is off
      The kill_switch is set to prevent us returning from a word at the
      top level. i.e. when the return stack is empty and we should be
      returning to CWeed Forths outer interpreter written in C.

Alg:  While kill switch not set
         The cell pointed to by ip is checked to see if its
         contents are the opcode of a Pinc primitive.
         If they are
            ip is moved onto the next cell
            the primitive is executed.
         Else
            the address of the next cell is pushed onto the return stack
         ip is set to the contents of the cell

       Pinc Version.
       We run the interpreter forever.

Alg:  Forever:
         The cell pointed to by ip is checked to see if its
         contents are the opcode of a Pinc primitive.
         If they are
            ip is moved onto the next cell
            the primitive is executed.
         Else
            the address of the next cell is pushed onto the return stack
         ip is set to the contents of the cell


**********************************************************************/

void inner_interpreter( void )
{
register unsigned CELL token;

#ifdef CWEED
    while( !kill_switch ) {
#ifdef CHECK_INNER
     if( is_primitive( ip )) {
         do_error( (CELL) 158 );
         longjmp( environment, 254 );
     }
#endif
     token = *((unsigned CELL *) &ram[ ip ]);
     ip += CELL_SIZE;
     if( is_primitive( token )) {
          (execution_vector[ (int) token ])();
     }
     else {
          push_return( ip );
          ip = token;
     }

    }

#else
again:
#ifdef CHECK_INNER
     if( is_primitive( ip )) {
         do_error( (CELL) 158 );
         longjmp( environment, 254 );
     }
#endif
     token = *((unsigned CELL *) &ram[ ip ]);
     ip += CELL_SIZE;
     if( is_primitive( token )) {
          (execution_vector[ (int) token ])();
     } else {
      /*  push_return( ip );  */
          return_stack[ ++rp ] = (unsigned CELL) ip;
          ip = token;
     }
    goto again;
#endif
}





/******************************  execute  *************************

Desc: Executes the word whose cfa is given.

In:   cfa  address of the word to be executed.

Alg:  If we are running CWeed Forth the kill switch is turned off
      since we do wish to execute at least this word.
      ip is set to the cfa and we call the inner interpreter

*******************************************************************/

void execute( cfa )
unsigned CELL cfa;
{
#ifdef CWEED
    kill_switch = 0;
#endif
    ip = cfa;
    inner_interpreter();
}

