#makes main

CC=cc -g -m64

#Fidl Compiler
#
# The Fidl Compiler does not depend on any other files.
# It generates the following header files
#     ENUM.H, PROTO.H, EXTPROTO.H, PRIM.H, EXEC.H
# Thus any file which imports one or more of these
# headers is dependant upon the Fidl compiler and
# the Fidl.src file.

all:	pinc cweed fidl lexicon

fidl.o:	fidl.c
	$(CC) -c fidl.c

fidl:	fidl.o
	$(CC) -o fidl fidl.o;


prim.h:	fidl.src fidl
	./fidl fidl.src


casecon.o:	casecon.c
	$(CC) -c casecon.c


io.o:	io.c pinc.h cmodel.h vmsize.h lowram.h strsizes.h
	$(CC) -c io.c


uxsignal.o:	uxsignal.c
	$(CC) -c uxsignal.c


headtrav.o:	headtrav.c pinc.h cmodel.h vmsize.h nfmasks.h enum.h
	$(CC) -c headtrav.c


threads.o:	threads.c pinc.h vocs.h cmodel.h lowram.h misc.h\
	     vmsize.h nfmasks.h
	$(CC) -c threads.c


prim.o:	prim.c pinc.h misc.h cmodel.h vmsize.h lowram.h vocs.h\
	  strsizes.h struct.h nfmasks.h prim.h proto.h os.h enum.h
	$(CC) -c prim.c


cwprim.o:	prim.c pinc.h misc.h cmodel.h vmsize.h lowram.h vocs.h\
	  strsizes.h struct.h nfmasks.h prim.h proto.h os.h enum.h
	$(CC) -DCWEED -c -o cwprim.o prim.c


error.o:	error.c extproto.h pinc.h cmodel.h\
          lowram.h strsizes.h fname.h vmsize.h
	$(CC) -c error.c


stacks.o:	stacks.c cmodel.h vmsize.h misc.h
	$(CC) -c stacks.c


inner.o:	inner.c pinc.h cmodel.h vmsize.h prim.h misc.h enum.h
	$(CC) -c inner.c


cwinner.o:	inner.c pinc.h cmodel.h vmsize.h prim.h enum.h
	$(CC) -DCWEED -c -o cwinner.o inner.c


carve.o:	carve.c cmodel.h dicmanip.c vocs.h lowram.h nfmasks.h vmsize.h\
	   prim.h misc.h
	$(CC) -c carve.c


dicmanip.o:	dicmanip.c pinc.h cmodel.h vmsize.h
	$(CC) -c dicmanip.c


dicload.o:	dicload.c pinc.h cmodel.h vmsize.h strsizes.h
	$(CC) -c dicload.c


n2str.o:	n2str.c cmodel.h vmsize.h lowram.h
	$(CC) -c n2str.c


str2n.o:	str2n.c cmodel.h vmsize.h lowram.h
	$(CC) -c str2n.c


cweed.o:	cweed.c pinc.h cmodel.h lowram.h nfmasks.h vmsize.h
	$(CC) -c cweed.c


main.o:	main.c prim.h pinc.h cmodel.h lowram.h vmsize.h fname.h enum.h
	$(CC) -c main.c


cwforth.o:	main.c prim.h pinc.h cmodel.h lowram.h vmsize.h
	$(CC) -DCWEED -c -o cwforth.o main.c


cweed:	cwforth.o threads.o cweed.o cwprim.o stacks.o\
	   cwinner.o carve.o dicmanip.o casecon.o\
	   dicload.o headtrav.o n2str.o str2n.o error.o io.o uxsignal.o
	$(CC) -o cweed cwforth.o threads.o cweed.o cwprim.o stacks.o \
	 cwinner.o carve.o dicmanip.o casecon.o\
	 dicload.o headtrav.o n2str.o str2n.o error.o io.o uxsignal.o


pinc:	cweed main.o threads.o prim.o stacks.o\
	    inner.o carve.o dicmanip.o casecon.o\
	    dicload.o headtrav.o n2str.o str2n.o error.o io.o uxsignal.o
	$(CC) -o pinc main.o threads.o prim.o stacks.o \
	 inner.o carve.o dicmanip.o casecon.o\
	 dicload.o headtrav.o n2str.o str2n.o error.o io.o uxsignal.o


lexicon:	prim.h misc.h cscape enum.h
	./cweed

clean:
	rm -f pinc cweed fidl lexicon *.o cscope.out
