/*******************************************************************

                Pinc Stack Operations

Author: Rod Crawford.

Date:  6/ 4/90

The following macros define the basic operations that can be perfomed
upon the data and return stacks. These macros replace the previous functions
for performing these operations.

************************************************************************/


extern unsigned CELL data_stack[ DATA_STACK_TOP ];
extern unsigned CELL return_stack[ RETURN_STACK_TOP ];
extern unsigned CELL sp, rp;




/**************************  push_data  ********************************

Desc: Pushes One CELL of data onto the data stack. The stack pointer
      is pre-decremented to point where the data is placed.

In: Data  The data item to be pushed on the stack.

***********************************************************************/

#define  push_data( data )    (data_stack[ ++sp ] = data)




/************************  pop_data  **********************************

Desc: Pops CELL from the top of data the stack and returns it.
      The stack pointer is post incremented to point at the next item
      in the stack.

Ret:  Data item popped from the stack.

**********************************************************************/

#define pop_data()    (data_stack[ sp-- ])




/**************************  push_return  ********************************

Desc: Pushes One address onto the return stack. The stack pointer
      is pre-decremented to point where the return is placed.

Args: Address  The item to be pushed onto the stack.

***********************************************************************/

#define push_return( address ) (return_stack[ ++rp ] = address)



/************************  pop_return  **********************************

Desc: Pops CELL from the top of return the stack and returns it.
      The stack pointer is post incremented to point at the next item
      in the stack.

Ret:  Item popped from the stack.

**********************************************************************/

#define pop_return()  (return_stack[ rp-- ])
