/********************************************************************

        Pinc Thread Table Functions

Author: Rod Crawford

Date: 16/11/89

Mods: 12/ 4/90 Rod Added support for OS9

This file contains the set of functions for traversing Pincs
Vocabulary Thread Tables

********************************************************************/

#include "cmodel.h"
#include "pinc.h"
#include "vocs.h"
#include "lowram.h"
#include "misc.h"
#include "vmsize.h"
#include "nfmasks.h"
#include "strsizes.h"
#include "nfmasks.h"



/**************************  external data *******************************/


#if IMPLEMENTATION == MicrosoftC
    extern unsigned char huge ram[ RAM_TOP ];
#else
#if IMPLEMENTATION == OS9
    extern remote unsigned char ram[ RAM_TOP ];
#else
    extern unsigned char ram[ RAM_TOP ];
#endif
#endif


extern unsigned CELL voc_link, dp;
extern struct from_fileblock {
     char filename[ FROM_FILENAME_LENGTH ];
     int page_number, line_number, line_in_page_number,
         last_page_required;
     FILE *stream;
       };


extern struct from_fileblock output;


/******************* External Function Prototypes  ************************/

#ifdef KNR
extern unsigned CELL lfa_to_nfa();
extern unsigned CELL nfa_to_lfa();
extern int hash_to_thread_number();
extern int hash_to_symbol_thread_number();
extern int smudged();
extern void strupper();
extern void strlower();
extern void set_raw_io();
extern void set_cooked_io();
#else
#warning "ANSI C  prototypes"
extern unsigned CELL lfa_to_nfa(unsigned CELL address);
extern unsigned CELL nfa_to_lfa( unsigned CELL address);
extern int hash_to_symbol_thread_number(char *name, unsigned CELL max_threads);
extern int smudged( unsigned CELL nfa );
extern void strupper( char *string );
extern void strlower( char *string );
#endif

/***************************  Function Prototypes  *********************/

#ifdef KNR
#warning "K and R prototypes"
unsigned CELL defined_word();
unsigned CELL defined_in_this_thread();
unsigned CELL defined_in_this_voc();
unsigned CELL defined_in_xc_vocabulary();
unsigned CELL defined_in_xc_thread();
void list_voc();
void copy_threads();
unsigned CELL sum();
unsigned CELL index_largest();
unsigned CELL current_voc();
void add_latest_voc_to_context();
int tabulate();
void forget_me();
unsigned CELL find_voc_below();
void trim_all_vocs_threads();
void trim_thread_table();
#else
#warning "ANS Defines"
unsigned CELL defined_word(char *word_name);
unsigned CELL defined_in_this_thread(char *word_name,
                         int name_length,

                         unsigned CELL *thread_link);
unsigned CELL defined_in_this_voc( char *word_name, unsigned CELL voc );
unsigned CELL defined_in_xc_vocabulary( char *word_name,
                         unsigned CELL thread_table,
                         unsigned CELL symbol_table_base );
unsigned CELL defined_in_xc_thread( char *word_name,
                        int name_length,
                        unsigned CELL *thread_link,
                        unsigned CELL symbol_table_base );
void list_voc(unsigned CELL voc);
void copy_threads(unsigned CELL *threads,unsigned CELL voc);
unsigned CELL sum(unsigned CELL *threads);
unsigned CELL index_largest(unsigned CELL *threads);
unsigned CELL current_voc(void );
void add_latest_voc_to_context(void );
int tabulate(char *name,int output_ptr);
void forget_me(unsigned CELL nfa);
unsigned CELL find_voc_below(unsigned CELL address );
void trim_all_vocs_threads(unsigned CELL address,unsigned CELL voc_link);
void trim_thread_table(unsigned CELL address,unsigned CELL voc_link);
#endif



/***************************  defined  *******************************

Desc: Searches each of the thread tables of the vocaubularies whose
      addresses are in the context area for the given word_name.
      Returns the lfa of the word if it exists otherwise 0.

In:   word_name string representing the word to be searched for.

Ret:  lfa of the word if found, else 0 (FALSE)

Alg:  Each vocabulary in the context area is searched until a
      match is found or they have all been searched.
      Searching commences form the first entry in context
      to the last (the function of a vocabulary is to add its cfa
      to the first item in the context array. Also shifts voc
      in the context up one cell to make space in the first cell)
      Searching involves hashing the name to gain an appropriate
      thread link. The thread link is then searched and if
      a match is found the lfa of the word is returned.
      If the thread becomes exhausted (*thread_link==0),
      searching continues at the next voc.
      If all the vocs have been exhausted 0 is returned.

NOTE: ATH
    Changed name from defined to defined_word because defined is
    a macro used by the pre-processor and searches across the
    source yield lots of hits.

    The pointre passed in now points to the length byte.
    This avoids situations like:

    03 A B C x 0

    The actual length of the name is 3, but strlen will report it as 4.

*********************************************************************/

unsigned CELL defined_word( word_name)
char word_name[];
{
    unsigned CELL * thread_link, voc, *context, lfa;
    int thread_number, i, name_length;
    unsigned char nlen;

//    nlen = word_name++;

#ifdef NO_CASE_SENSITIVITY
#ifdef USE_LOWER_CASE
     strlower( word_name );
#else
     strupper( word_name );
#endif
#else
#ifdef USER_LOWER_CASE
#if IMPLEMENTATION != SUN
   #error You must be insensitive to use lower case for Pinc
#endif
#endif
#endif
    // name_length = strlen( word_name );
    name_length = strlen( word_name);
    context = (unsigned CELL *) &ram[ CONTEXT_AREA ];
    thread_number = hash_to_thread_number( word_name );

    for( i=0; i < MAX_VOCS; i++ ) {
     if( (voc=(*(context+i))) ) {
         thread_link = ((unsigned CELL *) &ram[ voc ]) + 2 + thread_number;
//         if( (lfa=defined_in_this_thread( word_name, name_length, thread_link ))) {
         if( (lfa=defined_in_this_thread( word_name, name_length, thread_link ))) {
            return( lfa );
         }
     }
    }
    return( (CELL) 0 );
}


/***********************  defined_in_this_voc  ***********************

Desc: Searches the given vocabulary's thread table
      for the given word_name.
      Returns the lfa of the word if it exists otherwise 0.

      N.B. This is the same as the code of the loop body for
        "defined", however this function is not called by
        "defined" on gounds of efficientcy.

In:   word_name     string representing the word to be searched for.
      voc     vocabulary to be searched.
              note due to the new structure of PINC
              this value will not the cfa but the execution token.
              i.e. the cfa+1

Ret:  lfa of the word if found, else 0 (FALSE)

Alg:  The thread link is searched and if
      a match is found the lfa of the word is returned.
      Else 0 is returned.

**********************************************************************/

unsigned CELL defined_in_this_voc( word_name, voc )
char *word_name;
unsigned CELL voc;
{
    unsigned CELL * thread_link;
    int name_length;


#ifdef NO_CASE_SENSITIVITY
#ifdef USE_LOWER_CASE
     strlower( word_name );
#else
     strupper( word_name );
#endif
#else
#ifdef USER_LOWER_CASE
#if IMPLEMENTATION != SUN
     #error You must be insensitive to use lower case for Pinc
#endif
#endif
#endif
    name_length = strlen( word_name );

    if( voc ) {
     thread_link = ((unsigned CELL *) &ram[ voc ]) + 2
                + hash_to_thread_number( word_name );
     return( defined_in_this_thread( word_name,
                         name_length, thread_link ));
    }
}



/***********************  defined_in_this_thread  *********************

Desc: Searches the given thread tables
      for the given word_name.
      Returns the lfa of the word if it exists otherwise 0.

In:   word_name     string representing the word to be searched for.
      thread_table  thread table of vocabulary to be searched.

Ret:  lfa of the word if found, else 0 (FALSE)

Alg:  The thread link is searched and if
      a match is found the lfa of the word is returned.
      Else 0 is returned.

*********************************************************************/

unsigned CELL defined_in_this_thread( word_name, name_length, thread_link )
char *word_name;
int name_length;
unsigned CELL *thread_link;
{
    unsigned CELL this_nfa;

    while(  *thread_link ) {
     this_nfa = lfa_to_nfa((*thread_link));
     if( ((int) (ram[ this_nfa ] & (COUNT_BYTE_MASK)) == name_length)
         && !smudged( this_nfa )
         && !strcmp( word_name, &ram[ this_nfa+1 ] ) )
          return( *thread_link );
     thread_link = (unsigned CELL *) &ram[ *thread_link ];
    }
    return( 0 );
}




/***********************  defined_in_xc_vocabulary    ***********************

Desc: Searches the given symbol table's thread table
      for the given word_name.
      Returns the lfa of the word if it exists otherwise 0.

      N.B. This is the same as the code of the loop body for
        "defined", however this function is not called by
        "defined" on gounds of efficientcy.
        In addition, we have to add the address of the symbol
        table to each thread since all thread values are relative to
        this to facilitiate table relocation.

In:   word_name     string representing the word to be searched for.
      thread_table  thread table of vocabulary to be searched.

Ret:  lfa of the word if found, else 0 (FALSE)

Alg:  The thread link is searched and if
      a match is found the lfa of the word is returned.
      Else 0 is returned.

**********************************************************************/

unsigned CELL defined_in_xc_vocabulary( word_name,
                         thread_table, symbol_table_base )
char *word_name;
unsigned CELL thread_table, symbol_table_base;
{
    unsigned CELL *thread_link, *max_threads;
    int name_length;


#ifdef NO_CASE_SENSITIVITY
#ifdef USE_LOWER_CASE
     strlower( word_name );
#else
     strupper( word_name );
#endif
#else
#ifdef USER_LOWER_CASE
#if IMPLEMENTATION != SUN
     #error You must be insensitive to use lower case for Pinc
#endif
#endif
#endif
    name_length = strlen( word_name );
    max_threads = ((unsigned CELL *) &ram[ thread_table+symbol_table_base ]);
    thread_link = ((unsigned CELL *) &ram[ thread_table+symbol_table_base]) + 1
            + hash_to_symbol_thread_number( word_name, *max_threads );
    return( defined_in_xc_thread( word_name, name_length,
                      thread_link, symbol_table_base ));
}






/*********************defined_in_xc_thread  *********************

Desc: Searches the given thread
      for the given word_name.
      Returns the lfa of the word if it exists otherwise 0.
      Note since we are dealing with a relocatable symbol table,
      all thread links are relative to the symbol table offset. This must
      therefore be added in each time a thread is traversed

In:   word_name     string representing the word to be searched for.
      thread_table  thread table of vocabulary to be searched.

Ret:  lfa of the word if found, else 0 (FALSE)

Alg:  The thread link is searched and if
      a match is found the lfa of the word is returned.
      Else 0 is returned.

*********************************************************************/

unsigned CELL defined_in_xc_thread( word_name, name_length, thread_link,
                        symbol_table_base )
char *word_name;
int name_length;
unsigned CELL *thread_link, symbol_table_base;
{
    unsigned CELL this_nfa;


    while(  *thread_link ) {
     this_nfa = lfa_to_nfa((*thread_link) + symbol_table_base);
     if( ((int) (ram[ this_nfa ] & (COUNT_BYTE_MASK)) == name_length)
         && !smudged( this_nfa )
         && !strcmp( word_name, &ram[ this_nfa+1 ] ) )
          return( (*thread_link) + symbol_table_base );
     thread_link = (unsigned CELL *) &ram[ (*thread_link)
                               + symbol_table_base ];
    }
    return( 0 );
}







/******************************  list_voc  ***************************

Desc: Lists all the words in a given vocabulary. This word
      is used by words.

In:   voc  cfa address of the vocabulary to be listed.

Alg:  The threads in the vocabulary are copied to to a scratch thread
      table.
      While the thread table is not empty i.e. sum(table)!=0
            Find the largest address in the table (chronologically
            this must be the most recently compiled word not listed yet)
            This address is a link field so convert it to the nfa and
            print out the words name.
            Replace the entry in the thread table with the next lfa
            in the chain.

*******************************************************************/

void list_voc( voc )
unsigned CELL voc;
{
    unsigned CELL threads[ MAX_THREADS ], largest, *cell;
    int output_ptr, quit_out=0;
    char *name;

    copy_threads( threads, voc );
    output_ptr = 0;
#if IMPLEMENTATION == SUN
    set_raw_io();
#endif
    while( sum( threads ) && !quit_out) {

     largest = index_largest( threads );
     name = (char *) &ram[ lfa_to_nfa( threads[largest] )+1 ];
     fprintf( output.stream, "%s", name );
        output_ptr = tabulate( name, output_ptr );
     fflush( output.stream );
        cell = (unsigned CELL *) &ram[ threads[ largest ] ];
     threads[ largest ] = *cell;
#if IMPLEMENTATION == MicrosoftC
     if( kbhit())
         if( getch() == ' ' ) quit_out = (getch() != ' ');
         else
          quit_out = 1;
#else
#if IMPLEMENTATION == Linux

#else
#if IMPLEMENTATION == Linux64
#else
#if IMPLEMENTATION == TurboC
     if( kbhit())
         if( getch() == ' ' ) quit_out = (getch() != ' ');
         else
          quit_out = 1;

#else
#if IMPLEMENTATION == SUN
        if( key_pressed()) {
           if( getc( stdin ) == ' ' ) {
                 while( !key_pressed());
                 quit_out = (getc( stdin ) != ' ');
              }
           else
              quit_out = 1;
        }
#else
#if IMPLEMENTATION == OS9

#else
#if IMPLEMENTATION == MinGW64

#else
    #error No Implementation Specified
#endif
#endif
#endif
#endif
#endif
#endif

#endif
    }
#if IMPLEMENTATION == SUN
    set_cooked_io();
#endif
}


/************************  copy_threads  *******************************

Desc:  Copies the given vocabularies thread table into the supplied
       scratch thread table.

In:    voc      Address of vocabulary whose threads are to be copied

Out:   threads  copy of vocs threads.

************************************************************************/

void copy_threads( threads, voc )
unsigned CELL *threads, voc;
{
    int i;
    unsigned CELL vocs_thread_table, *cell;

    vocs_thread_table = voc+(2*CELL_SIZE);
    cell = (unsigned CELL *) &ram[ vocs_thread_table ];

    for( i=0; i <MAX_THREADS; i++ )
     threads[ i ] = *(cell+i);
}

/**************************   sum  *********************************

Desc: Retruns the sum of all the addresses in a given thread table

In:   threads  The address of the thread table to be summed

Ret:  SDum of the threads in the table.

*******************************************************************/

unsigned CELL sum( threads )
unsigned CELL *threads;
{
    int i;
    unsigned CELL sumation=0;

    for( i=0; i<MAX_THREADS; i++ )
     sumation += threads[ i ];

    return( sumation );
}


/************************* index_largest  ********************************

Desc: Returns the index in the thread table of the largest address
      Note the table is declared as unsigned!

In:  threads  The thread table to be searched

Ret: index of largest item in the thread table.

************************************************************************/

unsigned CELL index_largest( threads )
unsigned CELL *threads;
{
    int i, largest_index;
    unsigned CELL largest_so_far=0;

    for( i=0; i<MAX_THREADS; i++ )
     if( largest_so_far < threads[i] ) {
         largest_so_far = threads[i];
         largest_index = i;
      }
    return( largest_index );
}


/**************************  current_voc  *******************************

Desc:  Returns the cfa of the current vocabulary i.e. the one whose
       cfa is in the CURRENT_AREA

************************************************************************/

unsigned CELL current_voc( void )
{
    unsigned CELL *current;

    current = (unsigned CELL *) &ram[ CURRENT_AREA ];
    return( *current );
}



/***********************  add_latest_voc_to_context  *********************

Desc: Takes whatever the voc-link is pointing to and sticks it
      into the first cell of the CONTEXT_AREA

*************************************************************************/

void add_latest_voc_to_context( void )
{
    unsigned CELL * cell;

    cell = (unsigned CELL *) &ram[ CONTEXT_AREA ];
    *cell = voc_link;
}




/******************************  tabulate  ****************************

Desc: Ensures names are output in a tabular form. The given names
      length is evaulated and an additional number of spaces is
      printed to ensure future printing takes palce at the next
      tab stop.

In:   name  string to be printed
      output_ptr  current value ofthe output pointer

Ret:  new value of the output_ptr (set to next tab stop)

**********************************************************************/

int tabulate( name, output_ptr )
char *name;
int output_ptr;
{
    int i;

    if( (output_ptr + TAB_SIZE) > 64 ) {
        output_ptr = 0;
     fprintf( output.stream, "\n" );
    }
    else {
        for( i=0; i<( TAB_SIZE-(strlen( name ) % TAB_SIZE)); i++ )
         fprintf( output.stream, " " );
        output_ptr += TAB_SIZE;
    }
    return( output_ptr );
}



/**********************  forget_me  *********************************

Desc: forgets all words up to including the one represented
      by the given nfa

In:   nfa    name field address of word to be forgotten

Alg:  Find the first vocabulary below the nfa given.
      (Any vocabularies above this can be thrown away, the word
      cannot exist in them)
      If the vocabulary found is not the last one defined then
      we are going to be removing vocabularies. This implies
      that context and current may become ill-defined. They
      are therefore set to sensible values.
      All the vocabularies below the nfa have their thread tables
      trimmed. This involves chaining down each thread and removing
      all link which are greater than the nfa
      The dictioanry pointer is set to point at the lfa of
      the word we have just forgotten.

*******************************************************************/

void forget_me( nfa )
unsigned CELL nfa;
{
    unsigned CELL this_voc_link;
    this_voc_link = find_voc_below( nfa );
    if( this_voc_link != voc_link ) {
            only();
            definitions();
            voc_link = this_voc_link;
    }
    trim_all_vocs_threads( nfa_to_lfa( nfa ), voc_link );
    dp = nfa_to_lfa( nfa );
}



/***************************  find_voc_below  ****************************

Desc: Returns the cfa of the vocabulary that is less than or equal to
      the given address.

In:   address  address which vocabulary must be below

Ret:  the cfa of the vocabulary that is below the given address.

**************************************************************************/

unsigned CELL find_voc_below( address )
unsigned CELL address;
{
    unsigned CELL this_voc_link;

    this_voc_link = voc_link;
    while( address < this_voc_link )
         this_voc_link = next_voc_link( this_voc_link );
    return( this_voc_link );
}



/********************  trim_all_vocs_threads  **************************

Desc: Trims the threads of all the vocabularies that are below
      the given address removing all addresses in the threads that
      are greater than or equal to it.

In:   address  address that is to be trimmed out of the thread tables
               aCELL with all those above it.

     voc_link  initially a pointer to the cfa of the vocabulary immediately
               below the address, this variable is used to chain through
               all the vocabularies below the address.

***********************************************************************/

void trim_all_vocs_threads( address, voc_link )
unsigned CELL address, voc_link;
{

    while( voc_link ) {
     trim_thread_table( address, voc_link );
        voc_link = next_voc_link( voc_link );
    }
}




/********************  trim_thread_table  *******************************

Desc: Removes all items in the given vocabularies thread table
      that are above the given address.

In:   address address above which all items are to be removed
      voc_link  pointer to the cfa of the vocabulary whose
      thread table we are trimming

Alg:  for each entry in the thread table
            chain down the thread links until the address
         contained in a thread link is less than or equal
            to the address given.
            store the thread link value in the thread table.


**********************************************************************/

void trim_thread_table( address, voc_link )
unsigned CELL address, voc_link;
{
    unsigned CELL vocs_thread_table, *thread, *this_thread;
    int i;

    vocs_thread_table = voc_link+(2*CELL_SIZE);
    thread = (unsigned CELL *) &ram[ vocs_thread_table ];

    for( i=0; i<MAX_THREADS; i++ ) {
     this_thread = (thread+i);
     while( *this_thread >= address )
         this_thread = (unsigned CELL *) &ram[ *this_thread ];

     *(thread+i) = *this_thread;
    }
}
